#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef POSIX

#include <sys/socket.h>
#include <arpa/inet.h>

#else

#include "socket.h"
#include <conio.h>

#define htons(num) ((num>>8)|(num<<8))

void enable_break_sensing(void);
#pragma aux enable_break_sensing = \
  "mov ax, 0x3301" \
  "mov dl, 0x01" \
  "int 0x21" \
  modify [ax dx];

#endif

#include <unistd.h>
#include <sys/types.h>

struct disk_image
{
  FILE *fp;
  unsigned writable:1;
  char filename[13];
  unsigned long size;
};

unsigned disk_count = 0;

struct disk_image *disks;

int read_sector(unsigned disk_id, unsigned long offset, unsigned count, void *buff)
{
  fseek(disks[disk_id].fp, offset, 0);
  return fread(buff, count, 1, disks[disk_id].fp);
}

int write_sector(unsigned disk_id, unsigned long offset, unsigned count, void *buff)
{
  if(offset + count > disks[disk_id].size)
  {
    puts("Error: Offset is too large!");
    return 0;
  }
  fseek(disks[disk_id].fp, offset, 0);
  return fwrite(buff, count, 1, disks[disk_id].fp);
}

int open_disk_image(char *filename, unsigned writable)
{
  unsigned n = disk_count;
  char *ptr;
  char *wrrd[] =
  {
    "read-only",
    "read-write"
  };
  unsigned u;
  char *unit[] =
  {
    "bytes",
    "kilobytes",
    "megabytes"
  };

  if(!disk_count)
  {
    disks = malloc(sizeof(struct disk_image));
    if(!disks) return 2;
    disk_count++;
  }
  else
  {
    disk_count++;
    disks = realloc(disks, disk_count*sizeof(struct disk_image));
    if(!disks) return 2;
  }

  if(ptr = strchr(filename, '\n'))
    *ptr = 0;

  disks[n].fp = fopen(filename, (disks[n].writable = writable) ? "r+b" : "rb");

  if(!disks[n].fp)
  {
    printf("Cannot open disk image %s\n", filename);
    return 1;
  }

  printf("Opened disk image %s in %s mode\n", filename, wrrd[disks[n].writable]);

  strcpy(disks[n].filename, filename);

  fseek(disks[n].fp, 0, SEEK_END);
  disks[n].size = ftell(disks[n].fp);

  for(u=0;u<sizeof(unit)/sizeof(char*)-1;u++)
    if(!(disks[n].size >> ((u+1)*10)+2))
      break;
  printf("Disk size: %lu %s\n", disks[n].size>>(u*10), unit[u]);

  return 0;
}

int load_settings(char *filename)
{
  const char readonly_file[] = "RDFILE=";
  const char rw_file[] = "RWFILE=";
  char buff[100];
  int rc;

  FILE *fp = fopen(filename, "r");
  if(!fp) return 1;

  while(fgets(buff, sizeof(buff), fp))
  {
    if(!memcmp(buff, readonly_file, sizeof(readonly_file)-1))
      rc = open_disk_image(&buff[sizeof(readonly_file)-1], 0);
    else if(!memcmp(buff, rw_file, sizeof(rw_file)-1))
      rc = open_disk_image(&buff[sizeof(rw_file)-1], 1);
    if(rc == 2)
      puts("Out of memory.");
    if(rc) return 1;
  }

  fclose(fp);

  return 0;
}

// [R/W] + filename + offset (64-bit) + read count (16-bit) + payload
#define MAX_PAYLOAD_LEN 512
#define REQUEST_HEADER_LEN 1+12+8+2
#define REQUEST_LENGTH REQUEST_HEADER_LEN+MAX_PAYLOAD_LEN

int parse_request(unsigned char *request, unsigned *disk_id, uint64_t *offset, uint16_t *count, unsigned char *action)
{
  char filename[13];
  unsigned n;

  *action = request[0];

  memcpy(filename, &request[1], 12);
  filename[12] = 0;

  for(n=disk_count;n--;)
  {
    if(!strcmp(filename, disks[n].filename))
      break;
  }
  if(n == -1)
  {
    printf("Error: Disk %s not loaded.\n", filename);
    return -1;
  }

  *offset = *(uint64_t*)&request[1+12];
  *count = *(uint16_t*)&request[1+12+8];
  *disk_id = n;

  if(*action == 'R')
    printf("Read %u bytes from offset %llu from %s\n", (unsigned)*count, (unsigned long long)*offset, disks[n].filename);
  else if(*action == 'W')
    printf("Write %u bytes to offset %llu to %s\n", (unsigned)*count, (unsigned long long)*offset, disks[n].filename);

  return 0;
}

// action, filename, offset, count of bytes, payload
#define ANSWER_HEADER_LEN 1+12+8+2
#define ANSWER_LENGTH ANSWER_HEADER_LEN+MAX_PAYLOAD_LEN

int answer_request(int client_socket, unsigned disk_id, unsigned long long offset, unsigned count, struct sockaddr_in client, unsigned action, void *payload)
{
  unsigned char buff[ANSWER_LENGTH];
  unsigned anscount;
  int rc;

  if(count > MAX_PAYLOAD_LEN)
    count = MAX_PAYLOAD_LEN;

  anscount = count;

  if(action == 'R')
  {
    if(read_sector(disk_id, offset, count, &buff[ANSWER_HEADER_LEN]) != 1)
    {
      puts("Error: Cannot read image file.");
      return 1;
    }
    count += ANSWER_HEADER_LEN;
  }
  else if(action == 'W')
  {
    if(!disks[disk_id].writable)
    {
      puts("Error: Disk is write protected.");
      return 2;
    }
    if(write_sector(disk_id, offset, count, payload) != 1)
    {
      puts("Error: Cannot write image file.");
      return 3;
    }
    count = ANSWER_HEADER_LEN;
  }

  buff[0] = action;
  strncpy(&buff[1], disks[disk_id].filename, 12);
  *(uint64_t*)&buff[1+12] = offset;
  *(uint16_t*)&buff[1+12+8] = anscount;

  rc = sendto(client_socket, buff, count, 0, (struct sockaddr*)&client, sizeof(struct sockaddr_in));

  if(rc == -1)
    puts("sendto() failed.");
  else
    puts("Request answered.");

  return 0;
}

int serve_client(int client_socket)
{
  int c;
  unsigned char buff[REQUEST_LENGTH];
  unsigned disk_id;
  uint64_t offset;
  uint16_t read_count;
  struct sockaddr_in client;
  int addrlen = sizeof(struct sockaddr_in);
  unsigned char action;
  const unsigned char errnodisk[] = "E_NODISK";
  const unsigned char errcantread[] = "E_CANTREAD";
  const unsigned char errwrprot[] = "E_WRITEPROTECT";
  const unsigned char errcantwrite[] = "E_CANTWRITE";

  c = recvfrom(client_socket, buff, sizeof(buff), 0, (struct sockaddr*)&client, &addrlen);
  if(c == -1)
  {
    puts("recvfrom() failed.");
  }
  if(c >= REQUEST_HEADER_LEN)
  {
    printf("Accepted request from a client %u.%u.%u.%u.\n", ((unsigned char*)&client.sin_addr.s_addr)[0], ((unsigned char*)&client.sin_addr.s_addr)[1],
                                                            ((unsigned char*)&client.sin_addr.s_addr)[2], ((unsigned char*)&client.sin_addr.s_addr)[3]);
    c = parse_request(buff, &disk_id, &offset, &read_count, &action);
    if(c)
      c = sendto(client_socket, errnodisk, sizeof(errnodisk), 0, (struct sockaddr*)&client, sizeof(struct sockaddr_in));
    else
    {
      c = answer_request(client_socket, disk_id, offset, read_count, client, action, &buff[REQUEST_HEADER_LEN]);
      switch(c)
      {
        case 1: // cannot read
          c = sendto(client_socket, errcantread, sizeof(errcantread), 0, (struct sockaddr*)&client, sizeof(struct sockaddr_in));
          break;
        case 2: // write protected
          c = sendto(client_socket, errwrprot, sizeof(errwrprot), 0, (struct sockaddr*)&client, sizeof(struct sockaddr_in));
          break;
        case 3: // cannot write
          c = sendto(client_socket, errcantwrite, sizeof(errcantwrite), 0, (struct sockaddr*)&client, sizeof(struct sockaddr_in));
          break;
      }
    }
    if(c == -1)
      puts("sendto() failed.");
  }
}

int server_loop(int s)
{
  struct sockaddr_in client;
  int c = sizeof(struct sockaddr_in);

  while(1)
  {
#ifndef POSIX
    if(kbhit())
      if(getch() == 27)
        break;
#endif
    serve_client(s);
  }
}

int start_server(unsigned port)
{
  int s = socket(AF_INET, SOCK_DGRAM, 0);
  struct sockaddr_in server;

  if(s == -1)
  {
    puts("Could not create a new socket");
    return 1;
  }

  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons(port);

  if(bind(s, (struct sockaddr*)&server, sizeof(server)))
  {
    printf("Bind to port %u failed.\n", port);
    puts(strerror(errno));
    return 1;
  }

  server_loop(s);
}

int main(int argc, char *argv[])
{
  int rc;
  unsigned port = 300;

#ifndef POSIX
  enable_break_sensing();
#endif

  rc = load_settings("server.ini");

  if(argc > 1) port = atoi(argv[1]);

  if(!rc)
    return start_server(port);
}
